%{

#include "main.h"

extern void yyerror(const char * s);
extern int yylex();

// Store the current function
struct symbol* symb_scope_function;

// Boolean if we are in a print call
uint8_t isPrintCall = 0;

// stack of the IF ELSE calls
struct labelStack *  condition_stack;
// stacks of the DOM
struct domStack * dom_stack;


int indent;
char tabulation[INDENT_BUF];

// Store the dimension or indexes when we call or declare an array
struct symbol** tabArrayDim = NULL;
// Store the length of the array
unsigned int arraySizeLength;

// Function to call if we try to do exit(1). She frees the memory unlike exit()
void exit_safely(){
    if(tabArrayDim!=NULL){
        free(tabArrayDim);
    }
    yylex_destroy(); 
    free_stack((void**)condition_stack,LBLSTACK);
    free_stack((void**)dom_stack,DOMSTACK);
    free_exit(1,CODE,SYMTAB,fname);
}

%}

%define parse.error verbose

%start PROG

%union {
    long int intval;
    float floatval;
    name_t strval;
    wardLabels label;
    name_64_t chain;

    struct {
        struct symbol* ptr;
    } ptrVal;

    struct {
        name_t name;
        unsigned int dim;
    } ptrNameDim;

    struct {
        struct symbol* ptr;
        uint8_t isEArray;
    } ptrValE;

    struct {
        struct symbol* ptr1;
        struct symbol* ptr2;
        name_t name;
        unsigned int dim;
    } ptrArray;

    struct {
        uint8_t opnum;
    } opaff;
    
};

%token prog
%token def
%token Main
%token loc
%token dom
%token rec

%token in
%token out
%token inout

%token <strval> ID1
%token <strval> ID2
%token <chain> chaine
%token <intval> entier
%token <floatval> flottant

%token print
%token printstr
%token read_f

%token left_curly_bracket
%token right_curly_bracket
%token left_bracket
%token right_bracket
%token left_parenthesis
%token right_parenthesis
%token semicolon
%token comma
%token plus_aff
%token sub_aff
%token mult_aff
%token div_aff
%token not_recognized_aff
%token aff
%token double_period
%token in_dom

%token ternary_then
%token ternary_else

%token equal
%token inf
%token sup
%token sup_equal
%token inf_equal

%token plus
%token mult
%token sub
%token divide
%token mod

%right aff plus_aff sub_aff mult_aff div_aff
%left plus sub
%left mult divide mod
%right usub formatmod

%type <ptrNameDim> DLIST ARRAY ELIST
%type <ptrValE> E T F CALL
%type <ptrArray> ARRAYREF 
%type <opaff> OPAFF OPREL




%%

PROG : 
    prog ID left_curly_bracket 
    LFONC // Definition of functions
    def Main // Definition of the main function
        {
            struct symbol* id = symtable_get(SYMTAB,"Main",symb_scope_function);
            if ( id != NULL ){
                fprintf(stderr, "Conflict at line %u: Function '%s' already declared.\n",lineNumber,"Main");
                exit_safely();
            }
            id = symtable_put_name(SYMTAB,"Main");
            symb_scope_function = id;
            gencode(CODE,DEF_MAIN,id,NULL,NULL);
        }
    left_curly_bracket LOC DOM REC right_curly_bracket right_curly_bracket
    ;

LFONC : 
    LFONC FONC 
    | %empty
    ;
FONC : 
    // Declaration of a function ID1
    def ID1 
        {
            struct symbol* id = symtable_get(SYMTAB,$2,symb_scope_function);
            if ( id != NULL ){
                fprintf(stderr, "Conflict at line %u: Function '%s' already declared.\n",lineNumber,$2);
                exit_safely();
            }
            id = symtable_put_name(SYMTAB,$2);
            symb_scope_function = id; // Change the current function

            gencode(CODE,DEF_FUNCTION,id,NULL,NULL);
        }
    left_curly_bracket ARGS LOC DOM REC right_curly_bracket
    ;

LOC : 
    loc LDECL 
    | loc 
    | %empty
    ;
LDECL : 
    LDECL comma DECL 
    | DECL
    ;
DECL : // Section Loc declaration of arguments
    ID1 
        {
            // Declare a variable without affectation
            struct symbol* id = symtable_get(SYMTAB,$1,symb_scope_function);
            if ( id != NULL ){
                fprintf(stderr, "Error at line %u in function '%s': Variable '%s' is already declared.\n",lineNumber,symb_scope_function->u.name,$1);
                exit_safely();
            }
            id = symtable_put_loc(SYMTAB,$1,symb_scope_function);
        }
    | ID1 aff E 
        {   
            // Declare a variable with affectation of value
            struct symbol* id = symtable_get(SYMTAB,$1,symb_scope_function);

            // Errors detection
            if($3.ptr->kind == CONSTANT_FLOAT){ // Float = array
                if(id == NULL){ // Test if the variable ID1 is already declare
                    fprintf(stderr, "Error at line %u in function '%s': Try to declare variable '%s' as a float (array of dimension and lenght 1) but the size isn't declare. You may prefer to do this : '(%s,1), '%s'=%0.2f'.\n",lineNumber,symb_scope_function->u.name,$1,$1,$1,$3.ptr->u.value_float);
                    exit_safely();
                }
                else if(id->kind != ARRAY){ // Test if the variable ID1 isn't an array (cannot affect an array value (float) to a non array variable)
                    fprintf(stderr, "Error at line %u in function '%s': Try assigning at variable '%s' a float value (an array of dimension and lenght 1) but '%s' isn't an array.\n",lineNumber,symb_scope_function->u.name,$1,$1);
                    exit_safely();
                }
            }
            else if($3.ptr->kind == ARRAY || $3.isEArray){ // Affectation is an array
                if(id == NULL){ // Test if the variable ID1 is already declare
                    fprintf(stderr, "Error at line %u in function '%s': Try to declare variable '%s' as an integer but the affectation value is an array.\n",lineNumber,symb_scope_function->u.name,$1);
                    exit_safely();
                }
                else if(id->kind != ARRAY){ // Test if ID1 isn't an array
                    fprintf(stderr, "Error at line %u in function '%s': Try assigning an array variable '%s' to a non array type variable '%s'.\n",lineNumber,symb_scope_function->u.name,$3.ptr->u.arr.name,$1);
                    exit_safely();
                }
            }
            else{ // Test if affectation value isn't a float and array
                if (id != NULL){ // Test if the variable ID1 is already declare
                    fprintf(stderr, "Error at line %u in function '%s': Variable '%s' is already declared.\n",lineNumber,symb_scope_function->u.name,$1);
                    exit_safely();
                }

                // Create a local variable name = ID1 in the scope of the 'symb_scope_function'
                id = symtable_put_loc(SYMTAB,$1,symb_scope_function); 
            }

            if(id->kind == ARRAY){ // Handle if ID1 is an implicit array(array -> array[0]) (ID1 is already declare, not by symtable_put_loc three instructions above)
                struct symbol* const_0 = symtable_const_int(SYMTAB,0);
                gencode(CODE,AFF_ARRAY,id,const_0,$3.ptr);
            }
            else // ID1 isn't an array
                gencode(CODE,COPY,id,$3.ptr,NULL);
        }
    | // Array declaration
      {
        // Initialization of variables, need to store the size of the different dimensions of the array
        arraySizeLength =3;
        tabArrayDim = (struct symbol**)malloc(arraySizeLength*sizeof(struct symbol*)); // Should be free after utilization and sets to NULL
        if(tabArrayDim==NULL){
            fprintf(stderr,"Memory allocation bug.\n");
            exit_safely();
        }
      }
      ARRAY
        {
            // Error detection
            struct symbol* id = symtable_get(SYMTAB,$2.name,symb_scope_function);
            if ( id != NULL ){
                fprintf(stderr, "Error at line %u in function '%s': Array '%s' is already declared.\n",lineNumber,symb_scope_function->u.name,id->u.arr.name);
                exit_safely();
            }
            if($2.dim==0){
                fprintf(stderr,"Bug array dimension equal 0\n");
                exit_safely();
            }

            // Compute the ND array length for a 1D array allocation (dim1*dim2*...*dimN with 3 addresses code)
            struct symbol* size=NULL;
            if($2.dim == 1){ //1D array doesn't need temporary operation
                size = tabArrayDim[0];
            } else{ //ND arrays, except 1D array, require temporary operation
                struct symbol* tmpPair;
                struct symbol* tmpImpair = newtemp(SYMTAB,INTEGER);
                gencode(CODE,BOP_MULT,tmpImpair,tabArrayDim[0],tabArrayDim[1]); // Init (case 2D array)
                
                for(unsigned int i=2;i<$2.dim;i++){
                    if(i%2 == 0){
                        tmpPair = newtemp(SYMTAB,INTEGER);
                        gencode(CODE,BOP_MULT,tmpPair,tmpImpair,tabArrayDim[i]);
                    } else{
                        tmpImpair = newtemp(SYMTAB,INTEGER);
                        gencode(CODE,BOP_MULT,tmpImpair,tmpPair,tabArrayDim[i]);
                    }                    
                }

                // If dimension is even -> last i was odd. If dimension is odd -> last i was even.
                ($2.dim%2 == 0)?(size = tmpImpair):(size = tmpPair);
            }
            
            // Create a variable of type array inside the scope of the 'symb_scope_function'
            struct symbol* arr = symtable_put_array(SYMTAB,$2.name,$2.dim,tabArrayDim,size,symb_scope_function);
            gencode(CODE,ALLOCATE_ARRAY,arr,NULL,NULL);

            free(tabArrayDim);
            tabArrayDim = NULL;
        }
    ;

ARRAY : // Array declaration shape (ID1, length_dim1, length_dim2, ..., length_dimN)
    left_parenthesis ID1 comma DLIST right_parenthesis
        {
            $$.dim = $4.dim;
            strcpy($$.name,$2);
        }
    ;
DLIST : // length_dim1, length_dim2, ..., length_dimN
    DLIST comma E
        {  
            // Error detection
            if($3.ptr->kind == CONSTANT_FLOAT){
                fprintf(stderr, "Error at line %u in function '%s': Try to allocate an array with a float length value.\n",lineNumber,symb_scope_function->u.name);
                exit_safely();
            }

            $$.dim +=1;
            if(arraySizeLength < $$.dim){//Realloc tabArrayDim if necessary
                arraySizeLength += 3;
                tabArrayDim=(struct symbol**)realloc(tabArrayDim,arraySizeLength*sizeof(struct symbol*));
            }
            tabArrayDim[$$.dim-1] = $3.ptr;
        }
    | E
        {
            if($1.ptr->kind == CONSTANT_FLOAT){
                fprintf(stderr, "Error at line %u in function '%s': Try to allocate an array with a float length value.\n",lineNumber,symb_scope_function->u.name);
                exit_safely();
            }
            $$.dim = 1;  // Call first by the parsing so we init here the number of dimension storage
            tabArrayDim[$$.dim-1] = $1.ptr;
        }
    ;

ARGS : 
    left_bracket SECTIONARGS LARGS right_bracket 
    | left_bracket right_bracket
    ;
LARGS : 
    LARGS semicolon SECTIONARGS 
    | %empty
    ;
SECTIONARGS : 
    in ARGLIST
    | out ARGLIST
    | inout ARGLIST
    ;
ARGLIST : 
    ARGLIST comma ELTARG
    | ELTARG
    ;
ELTARG : 
    ARRAY
    | ID1
    ;

DOM : 
    dom DOMLIST 
    | dom 
    | %empty
    ;
DOMLIST : 
    DOMLIST comma D 
    | D
    ;
D : 
    ID1 in_dom left_bracket E double_period E right_bracket
    {
        struct symbol * id = symtable_get(SYMTAB,$1,symb_scope_function);
        struct symbol * loop_start, * loop_end;
        
        if(id == NULL || strcmp(id->scope_function,symb_scope_function->u.name) != 0 )
        {
            id = symtable_put_loc(SYMTAB,$1,symb_scope_function);
        } else {
            fprintf(stderr, "Error at line %u in function '%s': DOM var shouldn't exist yet.\n",lineNumber,symb_scope_function->u.name);
            exit_safely();
        }
        
        loop_start = symtable_put_label(SYMTAB,"loop_start",SYMTAB->dom,symb_scope_function);
        loop_end = symtable_put_label(SYMTAB,"loop_end",SYMTAB->dom,symb_scope_function);
        ++(SYMTAB->dom);
        domQuad * to_push;
        CHK(to_push = malloc(sizeof(domQuad)),NULL);
        to_push->dom_var=id;
        to_push->lbl1=loop_start;
        to_push->lbl2=loop_end;
        // copy the start val into the dom var
        gencode(CODE,COPY,id,$4.ptr,NULL);
        // put the start loop label
        gencode(CODE,LABEL_W,loop_start,NULL,NULL);
        // we add an indent level since we enter a new label
        gencode(CODE, ADD_INDENT,NULL,NULL,NULL);
        // branch to end of loop if dom var (i,j,k or anything it's named) is greater than the end value of the loop (if lesser or equal is false)
        gencode(CODE,OPREL_LE,loop_end,id,$6.ptr);
        // now the start of the loop is initialized we just put the infos needed for the end in the stack
        push_stack((void*)(&dom_stack),to_push,DOMSTACK);
        
    }
    ;

REC : 
    rec BLOCKINST
    {
        // once the rec block is over we have to add the end of the different dom loops
        while(dom_stack != NULL)
        {
            // we get the head of the stack which should be the first elt since DOMLIST access the doms from right to left if wrongjust reverse it
            domQuad * tmp = head_stack((void**)(&dom_stack), DOMSTACK);
            // we can now gen the END LOOP code
            struct symbol * one = symtable_const_int(SYMTAB,1);
            // add one to the loop index
            gencode(CODE,BOP_PLUS,tmp->dom_var,tmp->dom_var,one);
            // jump to start of loop
            gencode(CODE,GOTO_LABEL,tmp->lbl1,NULL,NULL);
            // end of this loop
            gencode(CODE,RM_INDENT,NULL,NULL,NULL);
            // put the end loop label
            gencode(CODE,LABEL_W,tmp->lbl2,NULL,NULL);
            // no longer of use we delete it with it's elt
            pop_stack((void**)(&dom_stack), DOMSTACK, 1);
        }
    }
    | %empty // Error dectection -> Function without a rec section
        {
            fprintf(stderr, "Error at line %u in function '%s': Unexpected end of function (shouldn't be empty).\n",lineNumber,symb_scope_function->u.name);
            exit_safely();
        }
    ;
BLOCKINST : // Bloc of instructions
    left_curly_bracket LISTEI right_curly_bracket
    | I
    ;
LISTEI : 
    LISTEI semicolon I // Last instruction hasn't a semilocon after
    | I
    | %empty // Error detection -> Rec TODO section empty Rec isn't the only BLOCK using LISTEI
        {
            fprintf(stderr, "Error at line %u in function '%s': Unexpected end of rec section (shouldn't be empty).\n",lineNumber,symb_scope_function->u.name);
            exit_safely();
        }
    ;
I : 
    COND
    | AFFECTATION // Affectation of variables
    | CALL // Call of functions or variables
    | read_f left_parenthesis ID1 right_parenthesis // Read
            {
                struct symbol* id_read = symtable_get(SYMTAB,$3,symb_scope_function); 
                  
                gencode(CODE,CALL_READ,id_read,NULL,NULL);
            }
    | print left_parenthesis {isPrintCall = 1;} CALL right_parenthesis // Print variable (printing out var of functions ins't implement yet)
            {   
                gencode(CODE,CALL_PRINT,$4.ptr,NULL,NULL);
            }
    | print left_parenthesis PRECINST right_parenthesis // Print ?
    | printstr left_parenthesis chaine right_parenthesis // Prinstr
        {
            struct symbol* id = symtable_get(SYMTAB,$3,symb_scope_function);
            if ( id == NULL )
                id = symtable_put_chaine(SYMTAB,$3);

            gencode(CODE,CALL_PRINTSTR,id,NULL,NULL);
        }
    ;

OPAFF: 
    plus_aff     
        {
            $$.opnum = 1;
        }
    | sub_aff  
        {
            $$.opnum = 2;
        } 
    | mult_aff   
        {
            $$.opnum = 3;
        }  
    | div_aff   
        {
            $$.opnum = 4;
        }
    | not_recognized_aff
        {
            $$.opnum = 5;
        }
    ;

AFFECTATION :
    ARRAYREF OPAFF E // Assignement operation of E to an array element 
        {
            // Compute the Arrays [+-/*] E and store it in a temporary variable
            struct symbol* tmp1;
            struct symbol* tmp2;
            tmp1 = newtemp(SYMTAB,FLOAT);
            tmp2 = newtemp(SYMTAB,FLOAT);
            
            // Copy in a temporary var the value of ARRAYREF
            gencode(CODE,COPY_ARRAY,tmp1,$1.ptr1,$1.ptr2);

            // Find what assignement operation we are (OPAFF) and copy the operation in tmp2
            switch($2.opnum){
                case 1:
                    gencode(CODE,BOP_PLUS,tmp2,tmp1,$3.ptr);
                    break;

                case 2:
                    gencode(CODE,BOP_MINUS,tmp2,tmp1,$3.ptr);
                    break;

                case 3:
                    gencode(CODE,BOP_MULT,tmp2,tmp1,$3.ptr);
                    break;

                case 4:
                    gencode(CODE,BOP_DIV,tmp2,tmp1,$3.ptr);
                    break;

                case 5:
                    fprintf(stderr, "Error at line %u in function '%s': Unrecognized operation assignment (only +=, -=, *= and /= are possible).\n",lineNumber-1,symb_scope_function->u.name);
                    exit_safely();
                    break;

                default:
                    fprintf(stderr, "BUG\n");
                    break;
            }

            // Copy the tmp2 in the array memory
            gencode(CODE,AFF_ARRAY,$1.ptr1,$1.ptr2,tmp2);
        }
    | ID1 OPAFF E  // Assignement operation of E to ID1
        {
            struct symbol* tmpArr;
            struct symbol* id = symtable_get(SYMTAB,$1,symb_scope_function);
            struct symbol* t = NULL;
            // Errors detection
            if ( id == NULL ){
                fprintf(stderr, "Error at line %u in function '%s': Variable '%s' is not declared, neither as an argument, nor as a local variable.\n",lineNumber,symb_scope_function->u.name,$1);
                exit_safely();
            }
            else if(id->kind == ARRAY){ // Handle if ID1 is an implicit array(array -> array[0])
                if($3.ptr->kind == CONSTANT_INT){ // Cast
                    t = symtable_const_double(SYMTAB,(float)$3.ptr->u.value_int); 
                }
                else{
                    t = $3.ptr;
                }
                tmpArr = newtemp(SYMTAB,FLOAT);
                struct symbol* const_0 = symtable_const_int(SYMTAB,0);
                gencode(CODE,COPY_ARRAY,tmpArr,id,const_0);
            }
            else if($3.ptr->kind == CONSTANT_FLOAT){
                fprintf(stderr, "Error at line %u in function '%s': Try to affect a float to a non array variable '%s'.\n",lineNumber-1,symb_scope_function->u.name,$1);
                exit_safely();
            }
            else{
                t = $3.ptr;
                tmpArr = id;
            }

            // Compute the ID1 [+-/*] E and store it in a temporary variable
            enum var_type ty = INTEGER;
            if(ISFLOAT(t) || ISFLOAT(tmpArr))
                ty = FLOAT;
            struct symbol* tmp;
            
            tmp = newtemp(SYMTAB,ty);

            switch($2.opnum){
                case 1:
                    gencode(CODE,BOP_PLUS,tmp,tmpArr,t);
                    break;

                case 2:
                    gencode(CODE,BOP_MINUS,tmp,tmpArr,t);
                    break;

                case 3:
                    gencode(CODE,BOP_MULT,tmp,tmpArr,t);
                    break;

                case 4:
                    gencode(CODE,BOP_DIV,tmp,tmpArr,t);
                    break;

                case 5:
                    fprintf(stderr, "Error at line %u: Unrecognized operation assignment (only +=, -=, *= and /= are possible).\n",lineNumber-1);
                    exit_safely();
                    break;

                default:
                    fprintf(stderr, "BUG\n");
                    break;
            }

            // Copy the temp in the variable
            if(id->kind == ARRAY){ // Handle if ID1 is an implicit array(array -> array[0])
                struct symbol* const_0 = symtable_const_int(SYMTAB,0);
                gencode(CODE,AFF_ARRAY,id,const_0,tmp);
            }
            else
                gencode(CODE,COPY,id,tmp,NULL);
        }
    | ARRAYREF aff E // Affectation of E to an array element
        {               
            struct symbol* t = NULL;     

            if($3.ptr->kind == CONSTANT_INT){ // Cast
                float f = (float)$3.ptr->u.value_int;
                t = symtable_const_double(SYMTAB,f); 
            }
            else{
                t = $3.ptr;
            }
            //                      id      index  E 
            gencode(CODE,AFF_ARRAY,$1.ptr1,$1.ptr2,t);
        }
    | ID1 aff E  // Affectation of A to ID1
        {    
            struct symbol* id = symtable_get(SYMTAB,$1,symb_scope_function);
            
            // Errors detection
            if ( id == NULL ){
                fprintf(stderr, "Error at line %u in function '%s': Variable '%s' is not declared, neither as an argument, nor as a local variable.\n",lineNumber,symb_scope_function->u.name,$1);
                exit_safely();
            }

            if(id->kind != ARRAY && $3.isEArray){
                fprintf(stderr, "Error at line %u in function '%s': Try to affect a float to a non array variable '%s'.\n",lineNumber,symb_scope_function->u.name,$1);
                exit_safely();
            }

            if(id->kind == ARRAY){ // Handle if ID1 is an implicit array(array -> array[0])
                struct symbol* t;
                if($3.ptr->kind == CONSTANT_INT){ // Cast
                    t = symtable_const_double(SYMTAB,(float)$3.ptr->u.value_int); 
                }
                else{
                    t = $3.ptr;
                }
                struct symbol* const_0 = symtable_const_int(SYMTAB,0);
                gencode(CODE,AFF_ARRAY,id,const_0,t);
            }
            else
                gencode(CODE,COPY,id,$3.ptr,NULL);
        }
    ;

CALL : 
    ID1 left_parenthesis LCALL right_parenthesis // Call of functions
        {   
            $$.ptr = NULL;
        }
    | ID1    // Call variables
        {
            struct symbol* id = symtable_get(SYMTAB,$1,symb_scope_function);
            if ( id == NULL ){
                fprintf(stderr, "Error at line %u in function '%s': Variable '%s' is not declared, neither as an argument, nor as a local variable.\n",lineNumber,symb_scope_function->u.name,$1);
                exit_safely();
            }

            if(id->kind == ARRAY && !isPrintCall){ // If ID1 is an implicit array and we isn't in a print, we add [0] after ID1 : ID1[0]
                $$.isEArray = 1;
                $$.ptr = newtemp(SYMTAB,FLOAT);
                struct symbol* const_0 = symtable_const_int(SYMTAB,0);
                gencode(CODE,COPY_ARRAY,$$.ptr,id,const_0);
            }
            else // Else if we are in a print with an array or we just a normal variable, we just call ID1
                $$.ptr = id;
            
            isPrintCall = 0;
        }
    ;
LCALL : 
    LCALL comma CALL 
    | LCALL comma PRECINST 
    | CALL 
    | PRECINST
    ;

PRECINST : 
    mod entier %prec formatmod
    ;
ARRAYREF : // Array indexing
    ID1 left_bracket ELIST right_bracket // ID1[...]
        {
            // $$.ptr1 -> The ID1 symbol 
            // $$.ptr2 -> The index symbol

            struct symbol* id = symtable_get(SYMTAB,$1,symb_scope_function);
            if ( id == NULL){
                fprintf(stderr, "Error at line %u in function '%s': Array '%s' is not declared, neither as an argument, nor as a local array.\n",lineNumber,symb_scope_function->u.name,$1);
                exit_safely();
            }
            if(id->kind != ARRAY){
                fprintf(stderr, "Error at line %u in function '%s': Array '%s' is not declared as a local array.\n",lineNumber,symb_scope_function->u.name,$1);
                exit_safely();
            }
            if($3.dim==0){
                fprintf(stderr,"Bug array dimension equal 0\n");
                exit_safely();
            }
            if($3.dim!=id->u.arr.dim){
                fprintf(stderr,"Error at line %u in function '%s': Try to indexing %d dimensional array '%s' with %d index.\n",lineNumber,symb_scope_function->u.name,id->u.arr.dim,$1,$3.dim);
                exit_safely();
            }


            // Compute the ND array indexes to one index for an 1D array
            struct symbol* index;
            struct symbol* kIndex1Dpair;
            struct symbol* kIndex1Dimpaire;

            index=newtemp(SYMTAB,INTEGER);
            for(unsigned int k=0;k<$3.dim;k++){
                // Get K+1dim index
                struct symbol* kIndexND = tabArrayDim[k];

                // Error detection
                if(kIndexND->kind==CONSTANT_INT && id->u.arr.size[k]->kind==CONSTANT_INT){
                    if(kIndexND->u.value_int >= id->u.arr.size[k]->u.value_int){
                        fprintf(stderr, "Error at line %u in function '%s': Array '%s' can not be accessed with a number greater than the size of his dimension %d\n",lineNumber,symb_scope_function->u.name,$1,k+1);
                        exit_safely();
                    }
                }

                // Compute shift
                struct symbol* shift;
                if(k == $3.dim-1){
                    shift = NULL;
                } else if(k+1 == $3.dim-1){
                    shift = id->u.arr.size[k+1];
                } else{
                    struct symbol* tmpPair;
                    struct symbol* tmpImpair;
                    if(k%2==0){
                        tmpPair = newtemp(SYMTAB,INTEGER);
                        tmpImpair;
                        gencode(CODE,BOP_MULT,tmpPair,id->u.arr.size[k+1],id->u.arr.size[k+2]);
                    } else{
                        tmpPair;
                        tmpImpair = newtemp(SYMTAB,INTEGER);
                        gencode(CODE,BOP_MULT,tmpImpair,id->u.arr.size[k+1],id->u.arr.size[k+2]);
                    }
                    for(unsigned int j=k+3;j<$3.dim;j++){
                        id->u.arr.size[j];
                        if(j%2 == 0){
                            tmpPair = newtemp(SYMTAB,INTEGER);
                            gencode(CODE,BOP_MULT,tmpPair,tmpImpair,id->u.arr.size[j]);
                        } else{
                            tmpImpair = newtemp(SYMTAB,INTEGER);
                            gencode(CODE,BOP_MULT,tmpImpair,tmpPair,id->u.arr.size[j]);
                        }   
                    }

                    (($3.dim-1)%2==0)?(shift = tmpPair):(shift = tmpImpair);
                }

                // Compute the addition of the last iteration with the current
                if(k%2==0){ // Attention : shift can be NULL
                    if(k==0){ // If k equal 0, there isn't a previous iteration
                        if(shift!=NULL){
                            gencode(CODE,BOP_MULT,index,kIndexND,shift);
                        } else{
                            gencode(CODE,COPY,index,kIndexND,NULL);
                        }
                    } else{
                        kIndex1Dpair = newtemp(SYMTAB,INTEGER);
                        if(shift!=NULL){
                            gencode(CODE,BOP_MULT,kIndex1Dpair,kIndexND,shift);
                        } else{
                            gencode(CODE,COPY,kIndex1Dpair,kIndexND,NULL);
                        }
                        gencode(CODE,BOP_PLUS,index,index,kIndex1Dpair);
                    }
                } else{
                    kIndex1Dimpaire = newtemp(SYMTAB,INTEGER);
                    if(shift!=NULL){
                        gencode(CODE,BOP_MULT,kIndex1Dimpaire,kIndexND,shift);
                    } else{
                        gencode(CODE,COPY,kIndex1Dimpaire,kIndexND,NULL);
                    }
                    gencode(CODE,BOP_PLUS,index,index,kIndex1Dimpaire);
                }
            }

            // Free memory
            free(tabArrayDim);
            tabArrayDim = NULL;
            if(index == NULL){
                fprintf(stderr,"Bug index NULL\n");
                exit_safely();
            }

            // Pass the array symbol and the index
            $$.ptr1 = id;
            $$.ptr2 = index;
        }
    ;

ELIST : // Indexing list
    ELIST comma E // [...,E,E,E,E]
        {
            // Error detection, array cannot be accessed with a float value
            if($3.ptr->kind == CONSTANT_FLOAT){
                fprintf(stderr, "Error at line %u in function '%s': Try to access an array with a float index.\n",lineNumber,symb_scope_function->u.name);
                exit_safely();
            }

            $$.dim +=1;
            if(arraySizeLength < $$.dim){//Realloc tabArrayDim if necessary
                arraySizeLength += 3;
                tabArrayDim=(struct symbol**)realloc(tabArrayDim,arraySizeLength*sizeof(struct symbol*));
            }
            tabArrayDim[$$.dim-1] = $3.ptr;
        }
    | E
        {
            if($1.ptr->kind == CONSTANT_FLOAT){
                fprintf(stderr, "Error at line %u in function '%s': Try to access an array with a float index.\n",lineNumber,symb_scope_function->u.name);
                exit_safely();
            }
            if(tabArrayDim==NULL){
                // Initialization of variables, need to store the index of the different dimensions of the array
                arraySizeLength = 3;
                tabArrayDim = (struct symbol**)malloc(arraySizeLength*sizeof(struct symbol*));
                if(tabArrayDim==NULL){
                    fprintf(stderr,"Memory allocation bug.\n");
                    exit_safely();
                }
            } else{
                fprintf(stderr,"Bug tabArrayDim not null\n");
                exit_safely();
            }

            $$.dim = 1;
            tabArrayDim[$$.dim-1] = $1.ptr;
        }
    ;
        
E :
    E plus T
        {
            enum var_type ty = INTEGER;  // by default result is an integer
            if(ISFLOAT($1.ptr) || ISFLOAT($3.ptr)) // check if one of the operand is a float
                ty = FLOAT;  // since one operand is a float the result is a float
            $$.ptr = newtemp(SYMTAB,ty);
            if($1.ptr->kind == CONSTANT_INT && ($3.ptr->kind == CONSTANT_FLOAT || $3.ptr->kind == ARRAY || $3.ptr->kind == NAME_FLOAT)){ // If a float is T, then need to cast E to a float to do float+float => $$ E the result is array
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$1.ptr->u.value_int); 
                gencode(CODE,BOP_PLUS,$$.ptr,t,$3.ptr); 
            } else if($3.ptr->kind == CONSTANT_INT && ($1.ptr->kind == CONSTANT_FLOAT || $1.ptr->kind == ARRAY)){ // If a float is E, then need to cast E to a float to do float+float => $$ E the result is array
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$3.ptr->u.value_int); 
                gencode(CODE,BOP_PLUS,$$.ptr,$1.ptr,t); 
            } else{ // E et T ne sont pas des flottants
                gencode(CODE,BOP_PLUS,$$.ptr,$1.ptr,$3.ptr); 
            }
        }
    | E sub T
        {
            enum var_type ty = INTEGER;  // by default result is an integer
            if(ISFLOAT($1.ptr) || ISFLOAT($3.ptr)) // check if one of the operand is a float
                ty = FLOAT;  // since one operand is a float the result is a float
            $$.ptr = newtemp(SYMTAB,ty);
            if($1.ptr->kind == CONSTANT_INT && ($3.ptr->kind == CONSTANT_FLOAT || $3.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$1.ptr->u.value_int); 
                gencode(CODE,BOP_MINUS,$$.ptr,t,$3.ptr); 
            } else if($3.ptr->kind == CONSTANT_INT && ($1.ptr->kind == CONSTANT_FLOAT || $1.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$3.ptr->u.value_int); 
                gencode(CODE,BOP_MINUS,$$.ptr,$1.ptr,t); 
            } else{
                gencode(CODE,BOP_MINUS,$$.ptr,$1.ptr,$3.ptr); 
            }
        }
    | sub E %prec usub
        {
            enum var_type ty = INTEGER;  // by default result is an integer
            if(ISFLOAT($2.ptr)) // check if the operand is a float
                ty = FLOAT;  // since the operand is a float the result is a float
            $$.ptr = newtemp(SYMTAB,ty);
            gencode(CODE,UOP_MINUS,$$.ptr,$2.ptr,NULL);
        }
    | T
        {
            $$.ptr = $1.ptr;
            $$.isEArray = $1.isEArray;
        }
    ;

T : 
    T mult F 
        {
            enum var_type ty = INTEGER;  // by default result is an integer
            if(ISFLOAT($1.ptr) || ISFLOAT($3.ptr)) // check if one of the operand is a float
                ty = FLOAT;  // since one operand is a float the result is a float
            $$.ptr = newtemp(SYMTAB,ty);
            if($1.ptr->kind == CONSTANT_INT && ($3.ptr->kind == CONSTANT_FLOAT || $3.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$1.ptr->u.value_int); 
                gencode(CODE,BOP_MULT,$$.ptr,t,$3.ptr); 
            } else if($3.ptr->kind == CONSTANT_INT && ($1.ptr->kind == CONSTANT_FLOAT || $1.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$3.ptr->u.value_int); 
                gencode(CODE,BOP_MULT,$$.ptr,$1.ptr,t); 
            } else{
                gencode(CODE,BOP_MULT,$$.ptr,$1.ptr,$3.ptr); 
            }            
        }
    | T divide F
        {
            enum var_type ty = INTEGER;  // by default result is an integer
            if(ISFLOAT($1.ptr) || ISFLOAT($3.ptr)) // check if one of the operand is a float
                ty = FLOAT;  // since one operand is a float the result is a float
            $$.ptr = newtemp(SYMTAB,ty);
            if($1.ptr->kind == CONSTANT_INT && ($3.ptr->kind == CONSTANT_FLOAT || $3.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$1.ptr->u.value_int); 
                gencode(CODE,BOP_DIV,$$.ptr,t,$3.ptr); 
            } else if($3.ptr->kind == CONSTANT_INT && ($1.ptr->kind == CONSTANT_FLOAT || $1.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$3.ptr->u.value_int); 
                gencode(CODE,BOP_DIV,$$.ptr,$1.ptr,t); 
            } else{
                gencode(CODE,BOP_DIV,$$.ptr,$1.ptr,$3.ptr); 
            }  
        }
    | T mod F 
        {
            enum var_type ty = INTEGER;  // by default result is an integer
            if(ISFLOAT($1.ptr) || ISFLOAT($3.ptr)) // check if one of the operand is a float
                ty = FLOAT;  // since one operand is a float the result is a float
            $$.ptr = newtemp(SYMTAB,ty);
            if($1.ptr->kind == CONSTANT_INT && ($3.ptr->kind == CONSTANT_FLOAT || $3.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$1.ptr->u.value_int); 
                gencode(CODE,BOP_MOD,$$.ptr,t,$3.ptr); 
            } else if($3.ptr->kind == CONSTANT_INT && ($1.ptr->kind == CONSTANT_FLOAT || $1.ptr->kind == ARRAY)){
                $$.isEArray = 1;
                struct symbol* t = symtable_const_double(SYMTAB,(float)$3.ptr->u.value_int); 
                gencode(CODE,BOP_MOD,$$.ptr,$1.ptr,t); 
            } else{
                gencode(CODE,BOP_MOD,$$.ptr,$1.ptr,$3.ptr); 
            }
        }
    | F
        {
            $$.ptr = $1.ptr; 
            $$.isEArray = $1.isEArray;
        }
    ;

F :
    left_parenthesis E right_parenthesis 
        {
            $$.ptr = $2.ptr;
        }
    | CALL 
        {
            $$.isEArray = $1.isEArray;
            $$.ptr = $1.ptr;
        }
    | ARRAYREF 
        {
                       
            $$.ptr = newtemp(SYMTAB,FLOAT);
            gencode(CODE,COPY_ARRAY,$$.ptr,$1.ptr1,$1.ptr2);
        }
    | entier 
        {
            $$.ptr = symtable_const_int(SYMTAB,$1); 
        }
    | flottant
        {
            $$.ptr = symtable_const_double(SYMTAB,$1); 
        }
    ;

COND : 
    CONDCHECK THEN B ELSE B
        {
            // ENDIF label
            gencode(CODE,RM_INDENT,NULL,NULL,NULL);
            gencode(CODE,LABEL_W,condition_stack->elt->lbl2,NULL,NULL);
            
            pop_stack((void**)(&condition_stack),LBLSTACK,1); // we pop the current condition's labels since the condition is now over
        }
    ;
B : 
    BLOCKINST 
    | %empty
    ;


OPREL : 
    inf 
        {
            $$.opnum = 1;
        }
    | sup 
        {
            $$.opnum = 2;
        }
    | inf_equal 
        {
            $$.opnum = 3;
        }
    | sup_equal 
        {
            $$.opnum = 4;
        }
    | equal
        {
            $$.opnum = 5;
        }
    ;
    
ID :
    ID1
    | ID2
    ;

CONDCHECK:
    E OPREL E  {

        
        wardLabels * tmp;
        CHK(tmp = malloc(sizeof(wardLabels)),NULL); 
        tmp->cond_number = SYMTAB->condition;
        tmp->lbl1 = symtable_put_label(SYMTAB, "else", tmp->cond_number, symb_scope_function);
        tmp->lbl2 = symtable_put_label(SYMTAB, "endif", tmp->cond_number, symb_scope_function);
        switch($2.opnum){
            case 1:
                gencode(CODE,OPREL_LT,tmp->lbl1,$1.ptr,$3.ptr);
                break;

            case 2:
                gencode(CODE,OPREL_GT,tmp->lbl1,$1.ptr,$3.ptr);
                break;

            case 3:
                gencode(CODE,OPREL_LE,tmp->lbl1,$1.ptr,$3.ptr);
                break;

            case 4:
                gencode(CODE,OPREL_GE,tmp->lbl1,$1.ptr,$3.ptr);
                break;

            case 5:
                gencode(CODE,OPREL_EQ,tmp->lbl1,$1.ptr,$3.ptr);
                break;

            default:
                fprintf(stderr, "BUG\n");
                break;
        }
        push_stack((void**)(&condition_stack),(void*)tmp,LBLSTACK);
        (SYMTAB->condition) ++;
        // go to else si la condition est fausse

    }
    ;

THEN : 
    ternary_then
    {
        
        struct symbol * if_label = symtable_put_label(SYMTAB, "if", condition_stack->elt->cond_number, symb_scope_function);
        // lavel de la sectrion IF
        gencode(CODE, LABEL_W, if_label, NULL, NULL);
        gencode(CODE,ADD_INDENT,NULL,NULL,NULL);
    }
    ;

ELSE :
    ternary_else
    {
        // goto endif juste après le code du IF
        gencode(CODE, GOTO_LABEL, condition_stack->elt->lbl2, NULL, NULL);
        gencode(CODE,RM_INDENT,NULL,NULL,NULL);
        // label de la section ELSE
        gencode(CODE, LABEL_W, condition_stack->elt->lbl1, NULL, NULL);
        gencode(CODE,ADD_INDENT,NULL,NULL,NULL);
    }
    ;
%%

void yyerror(const char * s)
{
    fprintf(stderr,"corec.y: %s\n",s);
}
