#include "main.h"

struct symtable* SYMTAB;
struct code* CODE;
FILE* fname;

void display_credits();

int main(int argc, char *argv[])
{
    /*FILE* */fname = NULL;
    uint8_t boolean_tos = 0;
    // Loop through the command-line arguments
    for(int i = 1; i < argc; i++){
        if(strcmp(argv[i], "-version") == 0){
            display_credits();
        }
        else if(strcmp(argv[i], "-tos") == 0){
            boolean_tos = 1;
        }
        else if(strcmp(argv[i], "-o") == 0){
            if(i + 1 < argc) {
                fname = fopen(argv[i + 1],"w+");
                i++;
            } else{
                fprintf(stderr, "%s: Error: No output file name provided.\n",argv[0]);
                exit(1);
            }
        }
        else{
            fprintf(stderr, "%s: Error: Argument '%s' is not recognized -> argument ignored.\n",argv[0],argv[i]);
        }
    }

    // If there isn't -o output, print the result in stdout
    if(fname == NULL){fname = stdout;}

    // Init
    SYMTAB = symtable_new();
    CODE = code_new();

    // Parse with yacc file
    int r = yyparse();

    // Display symbole table
    if(boolean_tos){symtable_dump(SYMTAB,fname);}

    // Generate intermediate code
    code_dump(CODE,fname);

    code_mips_dump(SYMTAB,CODE,fname);

    // Free memory
    symtable_free(SYMTAB);
    code_free(CODE);
    yylex_destroy();

    if(fname!=NULL){fclose(fname);}

    return r;
}


void display_credits(){
    printf("----------------------------------------------------------------------------\n");
    printf("\tCompiler for COREC language by\n");
    printf("\t\t - Audric BONNEAU\n");
    printf("\t\t\t |_ https://gitlab.unistra.fr/audric.bonneau\n");
    printf("\t\t - Constantin ANDERSSEN\n");
    printf("\t\t\t |_ https://git.unistra.fr/canderssen\n");
    printf("\t\t - Julien CLAVEL\n");
    printf("\t\t\t |_ https://git.unistra.fr/j.clavel\n");
    printf("\t\t - Jaenai RUGENGANDE-IHIMBAZWE\n");
    printf("\t\t\t |_ https://git.unistra.fr/rugengande\n");
    printf("----------------------------------------------------------------------------\n");
}